@@ -1,28 +1,31 @@
-# -*- coding: utf-8 -*-
 """Tests the xonsh environment."""
-from __future__ import unicode_literals, print_function
-import os
-import re
-import pathlib
+
 import datetime
-import itertools
+import os
+import pathlib
+import re
+from random import shuffle
 from tempfile import TemporaryDirectory
+from threading import Thread
+from time import sleep
 
 import pytest
 
-from xonsh.tools import always_true, DefaultNotGiven
-from xonsh.commands_cache import CommandsCache
 from xonsh.environ import (
     Env,
-    locate_binary,
-    default_env,
-    make_args_env,
+    InternalEnvironDict,
     LsColors,
-    default_value,
     Var,
+    default_env,
+    default_value,
+    locate_binary,
+    make_args_env,
+    xonsh_cache_dir,
+    xonsh_config_dir,
+    xonsh_data_dir,
 )
-
-from tools import skip_if_on_unix
+from xonsh.pytest.tools import skip_if_on_unix
+from xonsh.tools import DefaultNotGiven, always_true
 
 
 def test_env_normal():
@@ -75,6 +78,15 @@ def test_env_detype_no_dict():
     assert "YO" not in det
 
 
+def test_env_detype_all():
+    env = Env()
+    env._vars["DEFAULT"] = Var.with_default(1)
+    env._detyped, det = None, env.detype()
+    env._detyped, det_all = None, env.detype_all()
+    assert "DEFAULT" not in det
+    assert "DEFAULT" in det_all
+
+
 def test_histcontrol_none():
     env = Env(HISTCONTROL=None)
     assert isinstance(env["HISTCONTROL"], set)
@@ -150,6 +162,45 @@ def test_swap_exception_replacement():
     assert env["VAR"] == "original value"
 
 
+def test_thread_local_swap():
+    env = Env(a="orig")
+    iter_count = 10
+    num_threads = 4
+    success_variables = [False] * (num_threads + 1)
+
+    def loop(index, swapped_values=None):
+        if swapped_values:
+            if env["a"] != "orig":
+                success_variables[index] = False
+                return
+            env.set_swapped_values(swapped_values)
+        for _ in range(iter_count):
+            if env["a"] != "swapped":
+                success_variables[index] = False
+                break
+            with env.swap(a=index):
+                sleep(0.01)
+                if env["a"] == index:
+                    success_variables[index] = True
+                else:
+                    success_variables[index] = False
+                    break
+            sleep(0.01)
+
+    with env.swap(a="swapped"):
+        threads = [
+            Thread(target=loop, args=(i, env.get_swapped_values()))
+            for i in range(1, num_threads + 1)
+        ]
+        for t in threads:
+            t.start()
+        loop(0)
+    for t in threads:
+        t.join()
+
+    assert all(success_variables)
+
+
 @skip_if_on_unix
 def test_locate_binary_on_windows(xession):
     files = ("file1.exe", "FILE2.BAT", "file3.txt")
@@ -160,7 +211,6 @@ def test_locate_binary_on_windows(xession):
             with open(fpath, "w") as f:
                 f.write(fpath)
         xession.env.update({"PATH": [tmpdir], "PATHEXT": [".COM", ".EXE", ".BAT"]})
-        xession.commands_cache = CommandsCache()
         assert locate_binary("file1") == os.path.join(tmpdir, "file1.exe")
         assert locate_binary("file1.exe") == os.path.join(tmpdir, "file1.exe")
         assert locate_binary("file2") == os.path.join(tmpdir, "FILE2.BAT")
@@ -168,9 +218,8 @@ def test_locate_binary_on_windows(xession):
         assert locate_binary("file3") is None
 
 
-def test_event_on_envvar_change(xession):
-    env = Env(TEST=0)
-    xession.env = env
+def test_event_on_envvar_change(xession, env):
+    env["TEST"] = 0
     share = []
     # register
 
@@ -184,9 +233,7 @@ def test_event_on_envvar_change(xession):
     assert share == ["TEST", 0, 1]
 
 
-def test_event_on_envvar_new(xession):
-    env = Env()
-    xession.env = env
+def test_event_on_envvar_new(xession, env):
     share = []
     # register
 
@@ -200,9 +247,8 @@ def test_event_on_envvar_new(xession):
     assert share == ["TEST", 1]
 
 
-def test_event_on_envvar_change_from_none_value(xession):
-    env = Env(TEST=None)
-    xession.env = env
+def test_event_on_envvar_change_from_none_value(xession, env):
+    env["TEST"] = None
     share = []
     # register
 
@@ -217,9 +263,8 @@ def test_event_on_envvar_change_from_none_value(xession):
 
 
 @pytest.mark.parametrize("val", [1, None, True, "ok"])
-def test_event_on_envvar_change_no_fire_when_value_is_same(val, xession):
-    env = Env(TEST=val)
-    xession.env = env
+def test_event_on_envvar_change_no_fire_when_value_is_same(val, xession, env):
+    env["TEST"] = val
     share = []
     # register
 
@@ -233,9 +278,7 @@ def test_event_on_envvar_change_no_fire_when_value_is_same(val, xession):
     assert share == []
 
 
-def test_events_on_envvar_called_in_right_order(xession):
-    env = Env()
-    xession.env = env
+def test_events_on_envvar_called_in_right_order(xession, env):
     share = []
     # register
 
@@ -286,7 +329,7 @@ def test_delitem():
     env = Env(VAR="a value")
     assert env["VAR"] == "a value"
     del env["VAR"]
-    with pytest.raises(Exception):
+    with pytest.raises(KeyError):
         env["VAR"]
 
 
@@ -301,7 +344,7 @@ def test_delitem_default():
     assert env[a_key] == a_value
 
 
-def test_lscolors_target(xonsh_builtins):
+def test_lscolors_target(xession):
     lsc = LsColors.fromstring("ln=target")
     assert lsc["ln"] == ("RESET",)
     assert lsc.is_target("ln")
@@ -327,9 +370,9 @@ def test_lscolors_events(key_in, old_in, new_in, test, xession):
     @xession.builtins.events.on_lscolors_change
     def handler(key, oldvalue, newvalue, **kwargs):
         nonlocal old_in, new_in, key_in, event_fired
-        assert (
-            key == key_in and oldvalue == old_in and newvalue == new_in
-        ), "Old and new event values match"
+        assert key == key_in and oldvalue == old_in and newvalue == new_in, (
+            "Old and new event values match"
+        )
         event_fired = True
 
     xession.env["LS_COLORS"] = lsc
@@ -412,7 +455,7 @@ def test_register_custom_var_bool(val, converted):
         (32, "32"),
         (0, "0"),
         (27.0, "27.0"),
-        (None, "None"),
+        (None, ""),
         ("lol", "lol"),
         ("false", "false"),
         ("no", "no"),
@@ -437,7 +480,7 @@ def test_register_var_path():
     # Empty string is None to avoid uncontrolled converting empty string to Path('.')
     path = ""
     env["MY_PATH_VAR"] = path
-    assert env["MY_PATH_VAR"] == None
+    assert env["MY_PATH_VAR"] is None
 
     with pytest.raises(TypeError):
         env["MY_PATH_VAR"] = 42
@@ -536,6 +579,21 @@ def test_env_get_defaults():
     assert "TEST_REG_DNG" not in env
 
 
+def test_env_class_repr():
+    """Class with repr return string if env var."""
+
+    class Cls:
+        def __init__(self, var):
+            self.var = var
+
+        def __repr__(self):
+            return self.var
+
+    env = Env(CLS=Cls("hello"))
+    assert str(env.get("CLS")) == "hello"
+    assert str(env.__getitem__("CLS")) == "hello"
+
+
 @pytest.mark.parametrize(
     "val,validator",
     [
@@ -547,3 +605,58 @@ def test_env_get_defaults():
 def test_var_with_default_initer(val, validator):
     var = Var.with_default(val)
     assert var.validate.__name__ == validator
+
+
+def test_thread_local_dict():
+    d = InternalEnvironDict()
+    d["a"] = 1
+    assert d["a"] == 1
+    d.set_locally("a", 2)
+    assert d["a"] == 2
+    d.set_locally("a", 3)
+    assert d["a"] == 3
+    d["a"] = 4
+    assert d["a"] == 4
+    d.del_locally("a")
+    assert d["a"] == 1
+    d.set_locally("a", 5)
+    assert d.pop("a") == 5
+    assert d["a"] == 1
+    d.set_locally("a", 6)
+    assert d.popitem() == ("a", 6)
+    assert d["a"] == 1
+    assert d.pop("a", "nope") == 1
+    assert d.pop("a", "nope") == "nope"
+    assert "a" not in d
+
+
+def test_thread_local_dict_multiple():
+    d = InternalEnvironDict()
+    num_threads = 5
+    thread_values = [None] * num_threads
+
+    def thread(i):
+        d.set_locally("a", i**2)
+        sleep(0.1)
+        thread_values[i] = d["a"]
+        sleep(0.1)
+        d.del_locally("a")
+        sleep(0.1)
+
+    threads = [Thread(target=thread, args=(i,)) for i in range(num_threads)]
+    shuffle(threads)
+    for t in threads:
+        t.start()
+    for t in threads:
+        t.join()
+
+    assert thread_values == [i**2 for i in range(num_threads)]
+
+
+def test_xonsh_dir_vars():
+    env = Env(
+        XONSH_CONFIG_DIR="/config", XONSH_CACHE_DIR="/cache", XONSH_DATA_DIR="/data"
+    )
+    assert xonsh_config_dir(env), "/config"
+    assert xonsh_cache_dir(env), "/cache"
+    assert xonsh_data_dir(env), "/data"
